#include "Skill.h"

namespace armarx
{
    namespace skills
    {
        Skill::Skill(const SkillDescription& desc) : description(desc)
        {
            // replace constructor if you want to have a specific logging tag
            Logging::setTag("armarx::skills::" + description.skillId.toString());
        }

        // install a local condition via a lambda
        void
        Skill::installConditionWithCallback(std::function<bool()>&& f, std::function<void()>&& cb)
        {
            std::lock_guard l(conditionCallbacksMutex);
            conditionCallbacks.push_back({f, cb});
        }

        Skill::InitResult
        Skill::_init()
        {
            //ARMARX_IMPORTANT << "Initializing skill '" << description.skillName << "'";
            conditionCallbacks.clear();
            running = true;
            started = armarx::core::time::DateTime::Now();

            // install timeout condition
            installConditionWithCallback(
                [&]() {
                    return (armarx::core::time::DateTime::Now() >= (started + description.timeout));
                },
                [&]() { notifyTimeoutReached(); });

            conditionCheckingThread = std::thread(
                [&]()
                {
                    armarx::core::time::Metronome metronome(conditionCheckingThreadFrequency);
                    while (running) // when the skill ends/aborts this variable will be set to false
                    {
                        {
                            std::scoped_lock l(conditionCallbacksMutex);
                            for (auto& p : conditionCallbacks)
                            {
                                auto& f = p.first;
                                auto& cb = p.second;
                                if (f())
                                {
                                    cb();
                                }
                            }
                        }
                        const auto sleepDuration = metronome.waitForNextTick();
                        if (not sleepDuration.isPositive())
                        {
                            ARMARX_WARNING
                                << deactivateSpam() << "PeriodicSkill: execution took too long ("
                                << -sleepDuration << " vs "
                                << conditionCheckingThreadFrequency.toCycleDuration() << ")";
                        }
                    }
                });
            return {.status = TerminatedSkillStatus::Succeeded};
        }

        Skill::PrepareResult
        Skill::_prepare()
        {
            if (checkWhetherSkillShouldStopASAP())
            {
                return {.status = ActiveOrTerminatedSkillStatus::Aborted};
            }

            // Default nothing to prepare
            std::scoped_lock l(parametersMutex);
            if (not description.acceptedType)
            {
                return {.status = ActiveOrTerminatedSkillStatus::Succeeded};
            }
            if (this->parameters && this->parameters->fullfillsType(description.acceptedType))
            {
                // wait until parameters fulfill type
                return {.status = ActiveOrTerminatedSkillStatus::Succeeded};
            }

            // false if we have to wait for parameters
            return {.status = ActiveOrTerminatedSkillStatus::Running};
        }

        Skill::MainResult
        Skill::_main()
        {
            // Nothing here yet...
            return {.status = TerminatedSkillStatus::Succeeded};
        }

        Skill::ExitResult
        Skill::_exit()
        {
            // ARMARX_IMPORTANT << "Exiting Skill '" << description.skillName << "'";
            running = false; // stop checking conditions

            if (conditionCheckingThread.joinable())
            {
                conditionCheckingThread.join();
            }
            exited = armarx::core::time::DateTime::Now();
            return {.status = TerminatedSkillStatus::Succeeded};
        }

        Skill::InitResult
        Skill::initSkill()
        {
            auto _res = this->_init();
            auto res = this->init();
            return {.status = skills::MergeSkillStatus(_res.status, res.status)};
        }

        Skill::PrepareResult
        Skill::prepareSkill()
        {
            auto _res = this->_prepare();
            auto res = this->prepare();
            return {.status = skills::MergeSkillStatus(_res.status, res.status)};
        }

        Skill::MainResult
        Skill::mainOfSkill()
        {
            auto _res = this->_main();
            auto res = this->main();
            return {.status = skills::MergeSkillStatus(_res.status, res.status), .data = res.data};
        }

        Skill::ExitResult
        Skill::exitSkill()
        {
            auto res = this->exit();
            auto _res = this->_exit();
            return {.status = skills::MergeSkillStatus(_res.status, res.status)};
        }

        void
        Skill::notifyTimeoutReached()
        {
            timeoutReached = true;
            onTimeoutReached();
        }

        Skill::MainResult
        Skill::MakeAbortedResult(aron::data::DictPtr data)
        {
            return MainResult{
                .status = TerminatedSkillStatus::Aborted,
                .data = data,
            };
        }

        void
        Skill::notifySkillToStopASAP()
        {
            stopped = true;
            onStopRequested();
        }

        bool
        Skill::checkWhetherSkillShouldStopASAP() const
        {
            return stopped || timeoutReached;
        }

        // condition effects
        void
        Skill::onTimeoutReached()
        {
        }

        void
        Skill::onStopRequested()
        {
        }

        // always called before prepare (should not take longer than 100ms)
        Skill::InitResult
        Skill::init()
        {
            // Default nothing to init
            return {.status = TerminatedSkillStatus::Succeeded};
        }

        // always called before main (should not take longer than 100ms)
        Skill::PrepareResult
        Skill::prepare()
        {
            // Default nothing to prepare
            return {.status = ActiveOrTerminatedSkillStatus::Succeeded};
        }

        // always called after main or if skill fails (should not take longer than 100ms)
        Skill::ExitResult
        Skill::exit()
        {
            // Default nothing to exit
            return {.status = TerminatedSkillStatus::Succeeded};
        }

        Skill::MainResult
        Skill::main()
        {
            // This is just a dummy implementation
            ARMARX_IMPORTANT << "Dummy executing skill '" << description.skillId
                             << "'. Please overwrite this method.";
            return {.status = TerminatedSkillStatus::Succeeded, .data = nullptr};
        }
    } // namespace skills
} // namespace armarx
